/* 1. KEYWORDS ----------------------------------------------------------- */

#define elif          else if
#define acase         break; case
#define adefault      break; default
#define AGLOBAL               /* global (project-scope)                       */
#define EXPORT                /* global (project-scope)                       */
#define MODULE        static  /* external static (file-scope)                 */
#define PERSIST       static  /* internal static (function-scope)             */
#define AUTO          auto    /* automatic (function-scope)                   */
#define TRANSIENT     auto    /* automatic (function-scope)                   */

typedef signed char   FLAG;   /* 8-bit signed quantity (replaces BOOL)        */
typedef signed char   SBYTE;  /* 8-bit signed quantity (replaces Amiga BYTE)  */
typedef signed short  SWORD;  /* 16-bit signed quantity (replaces Amiga WORD) */
typedef signed long   SLONG;  /* 32-bit signed quantity (same as LONG)

2. DEFINES ------------------------------------------------------------ */

/* version-dependent constants */
#define TITLEBAR         "Worm Wars 8.22"
#define ABOUTSTRING      "Worm Wars 8.22 for Game Boy Advance"
#define VERSION          "$VER: Worm Wars 8.22 (29.7.2009)" /* always d.m.y format */
#define RELEASEDATE      "Monday 29 July 2009"
#define COPYRIGHT        " 1993-2009 Amigan Software"

#define CAUSEWAIT      100
#define DEFAULTLEVELS    5 // different for GBA
#define MAXLEVELS        5 // different for GBA
#define SAYLIMIT        60 /* in characters */

#define STDFIELDX       18 // different for GBA
#define STDFIELDY       11 // different for GBA
#define MAXFIELDX       STDFIELDX // different for GBA
#define MAXFIELDY       STDFIELDY // different for GBA
#define LEVELSIZE       ((STDFIELDX + 1) * (STDFIELDY + 1))
#define ARROWX          (fieldx + 1)
#define CENTREX         (STDFIELDX / 2)
#define CENTREY         (STDFIELDY / 2)
#define ENDXPIXEL       ((STARTXPIXEL + SQUAREX * (fieldx + 1)) - 1)
#define ENDYPIXEL       ((STARTYPIXEL + SQUAREY * (fieldy + 1)) - 1)

#define BONUS_TAIL     1
#define BONUS_GLOW     2
#define BONUS_TIME    10
#define BONUS_NUMBER  50
#define BONUS_LEVEL  100

/* samples */

#define FXBORN_DOG        0
#define FXBORN_RAIN       1
#define FXGET_LIGHTNING   2
#define FXBORN_MISSILE    3
#define FXBORN_PROTECTOR  4
#define FXDEATH_WORM      5
#define FXDO_ENCLOSE      6
#define FXBORN_FRAGMENT   7
#define FXGET_AMMO        8
#define FXGET_CYCLONE     9
#define FXGET_RAIN       10
#define FXGET_GROWER     11
#define FXGET_OBJECT     12
#define FXGET_POWERUP    13
#define FXGET_GRAVE      14
#define FXHELP           15
#define FXUSE_AMMO       16
#define FXUSE_ARMOUR     17
#define FXUSE_BOMB       18
#define FXUSE_TELEPORT   19
#define FXBORN_BIRD      20
#define FXPAIN           21 /* 21-24 */
#define FXAPPLAUSE       25 /* after each new hiscore */
#define FXGAMEOVER       26 /* if all worms are dead */
#define FXCLICK          27 /* rundown, keypresses */
#define FXRIFF           28 /* after each rundown */
#define FXSIREN          29 /* out of time */
#define FXDING           30
#define SAMPLES          FXDING

/* speeds */

#define VERYFAST          1
#define FAST              2
#define NORMAL            4
#define SLOW              8
#define VERYSLOW         16 /* must be non-0 */

/* The constant NORMAL is used for several purposes:
    a)  as a speed (therefore must not be VERYFAST, FAST, SLOW
        or VERYSLOW);
    b)  as a blitmode (therefore must not be BLACK or WHITE); and
    c)  as a pointer (therefore must not be anything that can be
        a brush).

objects... */
    #define AFFIXER              0
    #define AMMO                 1
    #define ARMOUR               2
    #define AUTOJUMP             3
    #define BONUS                4
    #define BRAKES               5
    #define CUTTER               6
    #define CYCLONE_O            7
    #define ENCLOSER             8
    #define GLOW                 9
    #define GROWER              10
    #define ICE                 11
    #define LIGHTNING           12
    #define MINIBOMB            13
    #define MINIHEALER          14
    #define MISSILE_O           15
    #define MULTIPLIER          16
    #define POWER               17
    #define PROTECTOR           18
    #define PULSE               19
    #define PUSHER              20
    #define REMNANTS            21
    #define SIDESHOT            22
    #define SLAYER              23
    #define SLOWER              24
    #define SUPERBOMB           25
    #define SUPERHEALER         26
    #define SWITCHER            27
    #define TREASURE            28
    #define UMBRELLA            29
    /* two spare numbers here */
/* field contents... */
    #define FIRSTNUMBER         30
    #define LASTNUMBER          38
    #define EMPTY               39
    #define SILVER              40
    #define GOLD                41
    #define DYNAMITE            42
    #define WOOD                43
    #define STONE               44
    #define METAL               45
    #define FROST               46
    #define SLIME               47
    #define TELEPORT            48
    #define ARROWUP             49
    #define ARROWDOWN           50
    #define START               51
/* colour sets... */
    #define FIRSTARROW          52 //  52- 55
    #define  LASTARROW          55
    #define FIRSTRAIN           56 //  56- 59
    #define  LASTRAIN           59
    #define FIRSTMISSILE        60 //  60- 63
    #define  LASTMISSILE        63
    #define FIRSTPROTECTOR      64 //  64- 67
    #define  LASTPROTECTOR      67
    #define FIRSTTAIL           68 //  68- 71
    #define  LASTTAIL           71
    #define FIRSTGLOW           72 //  72- 75
    #define  LASTGLOW           75
    #define FIRSTFIRE           76 //  76- 79
    #define  LASTFIRE           79
    #define FIRSTGRAVE          80 //  80- 83
    #define  LASTGRAVE          83
    #define FIRSTPAIN           84 //  84- 87
    #define  LASTPAIN           87
    #define FIRSTHEAD           88 //  88- 91
    #define  LASTHEAD           91
    #define FIRSTCHERRY         92 //  92- 95
    #define  LASTCHERRY         95
    #define FIRSTFLOWER         96 //  96- 99
    #define  LASTFLOWER         99
    #define FIRSTBANANA        100 // 100-103
    #define  LASTBANANA        103
/* never field contents... */
    #define MUSIC              104
    #define FX                 105
    #define ALL                106
    #define BLACKARROW         107
    #define BLACKENED          108
    #define WHITENED           109
    #define FIRSTBIRDFRAME     110
    #define LASTBIRDFRAME      113
    #define BULLRIGHT          114
    #define DOGAWAKENING       115
    #define DOGDORMANT         116
    #define FROGRIGHT          117
    #define FROGMOUTHLEFT      118
    #define FROGMOUTHRIGHT     119
    #define FROGTONGUE         120
    #define FIRSTMISSILEFRAME  121
    #define LASTMISSILEFRAME   140
    #define RABBITRIGHT        141
    #define FIRSTBANANAFRAME   142
    #define LASTBANANAFRAME    169
    #define CLOCK              170
    #define FIRSTPATTERN       171
    #define LASTPATTERN        182
    #define BOTH               183
    #define FIRSTREVNUMBER     184
    #define LASTREVNUMBER      192
/* creatures... */
    #define ANT                193
    #define BIRD               194
    #define BULL               195
    #define CAMEL              196
    #define CLOUD              197
    #define CYCLONE_C          198
    #define DOG                199
    #define FISH               200
    #define FRAGMENT           201
    #define FROG               202
    #define GIRAFFE            203
    #define KANGAROO           204
    #define MISSILE_C          205
    #define MONKEY             206
    #define MOUSE              207
    #define OCTOPUS            208
    #define ORB                209
    #define OTTER              210
    #define RABBIT             211
    #define RAIN               212
    #define SALAMANDER         213
    #define SNAIL              214
    #define SPIDER             215
    #define GOOSE              216
    #define HORSE              217
    #define BEAR               218
    #define ELEPHANT           219
    #define CHICKEN            220
    #define ZEBRA              221
    #define RHINOCEROS         222
    #define SNAKE              223
    #define PANDA              224
    #define ARRAYSIZE          PANDA

/* no imagery... */
    #define TEMPTAIL           (ARRAYSIZE +  1)
    #define TEMPWOOD           (ARRAYSIZE +  2)
    #define TEMPSLIME          (ARRAYSIZE +  3)
    #define TEMPLIGHTNING      (ARRAYSIZE +  4)
    #define TEMPSILVER         (ARRAYSIZE +  5)
    #define TEMPGOLD           (ARRAYSIZE +  6)
    #define NOSQUARE           (ARRAYSIZE +  7)
    #define BANGDYNAMITE       (ARRAYSIZE +  8)
    #define TEMPBANGDYNAMITE   (ARRAYSIZE +  9)
    #define TEMPDYNAMITE       (ARRAYSIZE + 10)

/* FIRSTOBJECT is always 0. */
/* LASTCREATURE is always ARRAYSIZE. */
    #define LASTOBJECT         UMBRELLA
    #define FIRSTEMPTY         EMPTY
    #define LASTEMPTY          GOLD
    #define FIRSTCREATURE      ANT
    #define LASTCREATURE       ARRAYSIZE
    #define BULLLEFT           BULL
    #define DOGCHASING         DOG
    #define FROGLEFT           FROG
    #define RABBITLEFT         RABBIT
    #define LETTERS            (LASTLETTER - FIRSTLETTER)
    #define HEAD               FIRSTHEAD
    #define BANANA             FIRSTBANANA
    #define ANYTHING           STONE /* should be something that is valid field contents */

/* heads */
    #define  GREENHEADUP     0
    #define    REDHEADUP     1
    #define   BLUEHEADUP     2
    #define YELLOWHEADUP     3
    #define  GREENHEADDOWN   4
    #define    REDHEADDOWN   5
    #define   BLUEHEADDOWN   6
    #define YELLOWHEADDOWN   7
    #define  GREENHEADLEFT   8
    #define    REDHEADLEFT   9
    #define   BLUEHEADLEFT  10
    #define YELLOWHEADLEFT  11
    #define  GREENHEADRIGHT 12
    #define    REDHEADRIGHT 13
    #define   BLUEHEADRIGHT 14
    #define YELLOWHEADRIGHT 15
    #define  GREENGLOWUP    16
    #define    REDGLOWUP    17
    #define   BLUEGLOWUP    18
    #define YELLOWGLOWUP    19
    #define  GREENGLOWDOWN  20
    #define    REDGLOWDOWN  21
    #define   BLUEGLOWDOWN  22
    #define YELLOWGLOWDOWN  23
    #define  GREENGLOWLEFT  24
    #define    REDGLOWLEFT  25
    #define   BLUEGLOWLEFT  26
    #define YELLOWGLOWLEFT  27
    #define  GREENGLOWRIGHT 28
    #define    REDGLOWRIGHT 29
    #define   BLUEGLOWRIGHT 30
    #define YELLOWGLOWRIGHT 31
    #define GREENHEAD_NW    32
    #define GREENHEAD_NE    33
    #define GREENHEAD_SW    34
    #define GREENHEAD_SE    35
    #define GREENGLOW_NW    36
    #define GREENGLOW_NE    37
    #define GREENGLOW_SW    38
    #define GREENGLOW_SE    39
    #define REDHEAD_NW      40
    #define REDHEAD_NE      41
    #define REDHEAD_SW      42
    #define REDHEAD_SE      43
    #define REDGLOW_NW      44
    #define REDGLOW_NE      45
    #define REDGLOW_SW      46
    #define REDGLOW_SE      47
    #define BLUEHEAD_NW     48
    #define BLUEHEAD_NE     49
    #define BLUEHEAD_SW     50
    #define BLUEHEAD_SE     51
    #define BLUEGLOW_NW     52
    #define BLUEGLOW_NE     53
    #define BLUEGLOW_SW     54
    #define BLUEGLOW_SE     55
    #define YELLOWHEAD_NW   56
    #define YELLOWHEAD_NE   57
    #define YELLOWHEAD_SW   58
    #define YELLOWHEAD_SE   59
    #define YELLOWGLOW_NW   60
    #define YELLOWGLOW_NE   61
    #define YELLOWGLOW_SW   62
    #define YELLOWGLOW_SE   63
    #define HEADARRAYSIZE   YELLOWGLOW_SE

/* tails */

    #define GN_START   0
    #define GG_START  29
    #define RN_START  58
    #define RG_START  87
    #define BN_START 116
    #define BG_START 145
    #define YN_START 174
    #define YG_START 203

    #define FIRSTMAGIC  29
    #define SECONDMAGIC (FIRSTMAGIC * 2)

    #define GN_E_NE  (GN_START +   0)
    #define GN_E_S   (GN_START +   1)
    #define GN_E_SE  (GN_START +   2)
    #define GN_NE_SE (GN_START +   3)
    #define GN_NE_SW (GN_START +   4)
    #define GN_NW_E  (GN_START +   5)
    #define GN_NW_NE (GN_START +   6)
    #define GN_NW_SE (GN_START +   7)
    #define GN_NW_SW (GN_START +   8)
    #define GN_N_E   (GN_START +   9)
    #define GN_N_NE  (GN_START +  10)
    #define GN_N_NW  (GN_START +  11)
    #define GN_N_S   (GN_START +  12)
    #define GN_N_SE  (GN_START +  13)
    #define GN_N_SW  (GN_START +  14)
    #define GN_N_W   (GN_START +  15)
    #define GN_SW_E  (GN_START +  16)
    #define GN_SW_SE (GN_START +  17)
    #define GN_S_NE  (GN_START +  18)
    #define GN_S_NW  (GN_START +  19)
    #define GN_S_SE  (GN_START +  20)
    #define GN_S_SW  (GN_START +  21)
    #define GN_W_E   (GN_START +  22)
    #define GN_W_NE  (GN_START +  23)
    #define GN_NW_W  (GN_START +  24)
    #define GN_W_S   (GN_START +  25)
    #define GN_W_SE  (GN_START +  26)
    #define GN_W_SW  (GN_START +  27)
    #define GN_ALL   (GN_START +  28)

    #define GG_E_NE  (GG_START +   0)
    #define GG_E_S   (GG_START +   1)
    #define GG_E_SE  (GG_START +   2)
    #define GG_NE_SE (GG_START +   3)
    #define GG_NE_SW (GG_START +   4)
    #define GG_NW_E  (GG_START +   5)
    #define GG_NW_NE (GG_START +   6)
    #define GG_NW_SE (GG_START +   7)
    #define GG_NW_SW (GG_START +   8)
    #define GG_N_E   (GG_START +   9)
    #define GG_N_NE  (GG_START +  10)
    #define GG_N_NW  (GG_START +  11)
    #define GG_N_S   (GG_START +  12)
    #define GG_N_SE  (GG_START +  13)
    #define GG_N_SW  (GG_START +  14)
    #define GG_N_W   (GG_START +  15)
    #define GG_SW_E  (GG_START +  16)
    #define GG_SW_SE (GG_START +  17)
    #define GG_S_NE  (GG_START +  18)
    #define GG_S_NW  (GG_START +  19)
    #define GG_S_SE  (GG_START +  20)
    #define GG_S_SW  (GG_START +  21)
    #define GG_W_E   (GG_START +  22)
    #define GG_W_NE  (GG_START +  23)
    #define GG_NW_W  (GG_START +  24)
    #define GG_W_S   (GG_START +  25)
    #define GG_W_SE  (GG_START +  26)
    #define GG_W_SW  (GG_START +  27)
    #define GG_ALL   (GG_START +  28)

    #define RN_E_NE  (RN_START +   0)
    #define RN_E_S   (RN_START +   1)
    #define RN_E_SE  (RN_START +   2)
    #define RN_NE_SE (RN_START +   3)
    #define RN_NE_SW (RN_START +   4)
    #define RN_NW_E  (RN_START +   5)
    #define RN_NW_NE (RN_START +   6)
    #define RN_NW_SE (RN_START +   7)
    #define RN_NW_SW (RN_START +   8)
    #define RN_N_E   (RN_START +   9)
    #define RN_N_NE  (RN_START +  10)
    #define RN_N_NW  (RN_START +  11)
    #define RN_N_S   (RN_START +  12)
    #define RN_N_SE  (RN_START +  13)
    #define RN_N_SW  (RN_START +  14)
    #define RN_N_W   (RN_START +  15)
    #define RN_SW_E  (RN_START +  16)
    #define RN_SW_SE (RN_START +  17)
    #define RN_S_NE  (RN_START +  18)
    #define RN_S_NW  (RN_START +  19)
    #define RN_S_SE  (RN_START +  20)
    #define RN_S_SW  (RN_START +  21)
    #define RN_W_E   (RN_START +  22)
    #define RN_W_NE  (RN_START +  23)
    #define RN_NW_W  (RN_START +  24)
    #define RN_W_S   (RN_START +  25)
    #define RN_W_SE  (RN_START +  26)
    #define RN_W_SW  (RN_START +  27)
    #define RN_ALL   (RN_START +  28)

    #define RG_E_NE  (RG_START +   0)
    #define RG_E_S   (RG_START +   1)
    #define RG_E_SE  (RG_START +   2)
    #define RG_NE_SE (RG_START +   3)
    #define RG_NE_SW (RG_START +   4)
    #define RG_NW_E  (RG_START +   5)
    #define RG_NW_NE (RG_START +   6)
    #define RG_NW_SE (RG_START +   7)
    #define RG_NW_SW (RG_START +   8)
    #define RG_N_E   (RG_START +   9)
    #define RG_N_NE  (RG_START +  10)
    #define RG_N_NW  (RG_START +  11)
    #define RG_N_S   (RG_START +  12)
    #define RG_N_SE  (RG_START +  13)
    #define RG_N_SW  (RG_START +  14)
    #define RG_N_W   (RG_START +  15)
    #define RG_SW_E  (RG_START +  16)
    #define RG_SW_SE (RG_START +  17)
    #define RG_S_NE  (RG_START +  18)
    #define RG_S_NW  (RG_START +  19)
    #define RG_S_SE  (RG_START +  20)
    #define RG_S_SW  (RG_START +  21)
    #define RG_W_E   (RG_START +  22)
    #define RG_W_NE  (RG_START +  23)
    #define RG_NW_W  (RG_START +  24)
    #define RG_W_S   (RG_START +  25)
    #define RG_W_SE  (RG_START +  26)
    #define RG_W_SW  (RG_START +  27)
    #define RG_ALL   (RG_START +  28)

    #define BN_E_NE  (BN_START +   0)
    #define BN_E_S   (BN_START +   1)
    #define BN_E_SE  (BN_START +   2)
    #define BN_NE_SE (BN_START +   3)
    #define BN_NE_SW (BN_START +   4)
    #define BN_NW_E  (BN_START +   5)
    #define BN_NW_NE (BN_START +   6)
    #define BN_NW_SE (BN_START +   7)
    #define BN_NW_SW (BN_START +   8)
    #define BN_N_E   (BN_START +   9)
    #define BN_N_NE  (BN_START +  10)
    #define BN_N_NW  (BN_START +  11)
    #define BN_N_S   (BN_START +  12)
    #define BN_N_SE  (BN_START +  13)
    #define BN_N_SW  (BN_START +  14)
    #define BN_N_W   (BN_START +  15)
    #define BN_SW_E  (BN_START +  16)
    #define BN_SW_SE (BN_START +  17)
    #define BN_S_NE  (BN_START +  18)
    #define BN_S_NW  (BN_START +  19)
    #define BN_S_SE  (BN_START +  20)
    #define BN_S_SW  (BN_START +  21)
    #define BN_W_E   (BN_START +  22)
    #define BN_W_NE  (BN_START +  23)
    #define BN_NW_W  (BN_START +  24)
    #define BN_W_S   (BN_START +  25)
    #define BN_W_SE  (BN_START +  26)
    #define BN_W_SW  (BN_START +  27)
    #define BN_ALL   (BN_START +  28)

    #define BG_E_NE  (BG_START +   0)
    #define BG_E_S   (BG_START +   1)
    #define BG_E_SE  (BG_START +   2)
    #define BG_NE_SE (BG_START +   3)
    #define BG_NE_SW (BG_START +   4)
    #define BG_NW_E  (BG_START +   5)
    #define BG_NW_NE (BG_START +   6)
    #define BG_NW_SE (BG_START +   7)
    #define BG_NW_SW (BG_START +   8)
    #define BG_N_E   (BG_START +   9)
    #define BG_N_NE  (BG_START +  10)
    #define BG_N_NW  (BG_START +  11)
    #define BG_N_S   (BG_START +  12)
    #define BG_N_SE  (BG_START +  13)
    #define BG_N_SW  (BG_START +  14)
    #define BG_N_W   (BG_START +  15)
    #define BG_SW_E  (BG_START +  16)
    #define BG_SW_SE (BG_START +  17)
    #define BG_S_NE  (BG_START +  18)
    #define BG_S_NW  (BG_START +  19)
    #define BG_S_SE  (BG_START +  20)
    #define BG_S_SW  (BG_START +  21)
    #define BG_W_E   (BG_START +  22)
    #define BG_W_NE  (BG_START +  23)
    #define BG_NW_W  (BG_START +  24)
    #define BG_W_S   (BG_START +  25)
    #define BG_W_SE  (BG_START +  26)
    #define BG_W_SW  (BG_START +  27)
    #define BG_ALL   (BG_START +  28)

    #define YN_E_NE  (YN_START +   0)
    #define YN_E_S   (YN_START +   1)
    #define YN_E_SE  (YN_START +   2)
    #define YN_NE_SE (YN_START +   3)
    #define YN_NE_SW (YN_START +   4)
    #define YN_NW_E  (YN_START +   5)
    #define YN_NW_NE (YN_START +   6)
    #define YN_NW_SE (YN_START +   7)
    #define YN_NW_SW (YN_START +   8)
    #define YN_N_E   (YN_START +   9)
    #define YN_N_NE  (YN_START +  10)
    #define YN_N_NW  (YN_START +  11)
    #define YN_N_S   (YN_START +  12)
    #define YN_N_SE  (YN_START +  13)
    #define YN_N_SW  (YN_START +  14)
    #define YN_N_W   (YN_START +  15)
    #define YN_SW_E  (YN_START +  16)
    #define YN_SW_SE (YN_START +  17)
    #define YN_S_NE  (YN_START +  18)
    #define YN_S_NW  (YN_START +  19)
    #define YN_S_SE  (YN_START +  20)
    #define YN_S_SW  (YN_START +  21)
    #define YN_W_E   (YN_START +  22)
    #define YN_W_NE  (YN_START +  23)
    #define YN_NW_W  (YN_START +  24)
    #define YN_W_S   (YN_START +  25)
    #define YN_W_SE  (YN_START +  26)
    #define YN_W_SW  (YN_START +  27)
    #define YN_ALL   (YN_START +  28)

    #define YG_E_NE  (YG_START +   0)
    #define YG_E_S   (YG_START +   1)
    #define YG_E_SE  (YG_START +   2)
    #define YG_NE_SE (YG_START +   3)
    #define YG_NE_SW (YG_START +   4)
    #define YG_NW_E  (YG_START +   5)
    #define YG_NW_NE (YG_START +   6)
    #define YG_NW_SE (YG_START +   7)
    #define YG_NW_SW (YG_START +   8)
    #define YG_N_E   (YG_START +   9)
    #define YG_N_NE  (YG_START +  10)
    #define YG_N_NW  (YG_START +  11)
    #define YG_N_S   (YG_START +  12)
    #define YG_N_SE  (YG_START +  13)
    #define YG_N_SW  (YG_START +  14)
    #define YG_N_W   (YG_START +  15)
    #define YG_SW_E  (YG_START +  16)
    #define YG_SW_SE (YG_START +  17)
    #define YG_S_NE  (YG_START +  18)
    #define YG_S_NW  (YG_START +  19)
    #define YG_S_SE  (YG_START +  20)
    #define YG_S_SW  (YG_START +  21)
    #define YG_W_E   (YG_START +  22)
    #define YG_W_NE  (YG_START +  23)
    #define YG_NW_W  (YG_START +  24)
    #define YG_W_S   (YG_START +  25)
    #define YG_W_SE  (YG_START +  26)
    #define YG_W_SW  (YG_START +  27)
    #define YG_ALL   (YG_START +  28)

    #define WHATEVER 232
    #define TAILARRAYSIZE WHATEVER

    #define GN_SE_NW      GN_NW_SE
    #define GN_SE_N       GN_N_SE
    #define GN_SE_NE      GN_NE_SE
    #define GN_SE_W       GN_W_SE
    #define GN_SE_E       GN_E_SE
    #define GN_SE_SW      GN_SW_SE
    #define GN_SE_S       GN_S_SE
    #define GN_S_N        GN_N_S
    #define GN_S_W        GN_W_S
    #define GN_S_E        GN_E_S
    #define GN_SW_NW      GN_NW_SW
    #define GN_SW_N       GN_N_SW
    #define GN_SW_NE      GN_NE_SW
    #define GN_SW_W       GN_W_SW
    #define GN_SW_S       GN_S_SW
    #define GN_E_NW       GN_NW_E
    #define GN_E_N        GN_N_E
    #define GN_E_W        GN_W_E
    #define GN_E_SW       GN_SW_E
    #define GN_W_NW       GN_NW_W
    #define GN_W_N        GN_N_W
    #define GN_NE_NW      GN_NW_NE
    #define GN_NE_N       GN_N_NE
    #define GN_NE_W       GN_W_NE
    #define GN_NE_E       GN_E_NE
    #define GN_NE_S       GN_S_NE
    #define GN_NW_N       GN_N_NW
    #define GN_NW_S       GN_S_NW

    #define GG_SE_NW      GG_NW_SE
    #define GG_SE_N       GG_N_SE
    #define GG_SE_NE      GG_NE_SE
    #define GG_SE_W       GG_W_SE
    #define GG_SE_E       GG_E_SE
    #define GG_SE_SW      GG_SW_SE
    #define GG_SE_S       GG_S_SE
    #define GG_S_N        GG_N_S
    #define GG_S_W        GG_W_S
    #define GG_S_E        GG_E_S
    #define GG_SW_NW      GG_NW_SW
    #define GG_SW_N       GG_N_SW
    #define GG_SW_NE      GG_NE_SW
    #define GG_SW_W       GG_W_SW
    #define GG_SW_S       GG_S_SW
    #define GG_E_NW       GG_NW_E
    #define GG_E_N        GG_N_E
    #define GG_E_W        GG_W_E
    #define GG_E_SW       GG_SW_E
    #define GG_W_NW       GG_NW_W
    #define GG_W_N        GG_N_W
    #define GG_NE_NW      GG_NW_NE
    #define GG_NE_N       GG_N_NE
    #define GG_NE_W       GG_W_NE
    #define GG_NE_E       GG_E_NE
    #define GG_NE_S       GG_S_NE
    #define GG_NW_N       GG_N_NW
    #define GG_NW_S       GG_S_NW

    #define RN_SE_NW      RN_NW_SE
    #define RN_SE_N       RN_N_SE
    #define RN_SE_NE      RN_NE_SE
    #define RN_SE_W       RN_W_SE
    #define RN_SE_E       RN_E_SE
    #define RN_SE_SW      RN_SW_SE
    #define RN_SE_S       RN_S_SE
    #define RN_S_N        RN_N_S
    #define RN_S_W        RN_W_S
    #define RN_S_E        RN_E_S
    #define RN_SW_NW      RN_NW_SW
    #define RN_SW_N       RN_N_SW
    #define RN_SW_NE      RN_NE_SW
    #define RN_SW_W       RN_W_SW
    #define RN_SW_S       RN_S_SW
    #define RN_E_NW       RN_NW_E
    #define RN_E_N        RN_N_E
    #define RN_E_W        RN_W_E
    #define RN_E_SW       RN_SW_E
    #define RN_W_NW       RN_NW_W
    #define RN_W_N        RN_N_W
    #define RN_NE_NW      RN_NW_NE
    #define RN_NE_N       RN_N_NE
    #define RN_NE_W       RN_W_NE
    #define RN_NE_E       RN_E_NE
    #define RN_NE_S       RN_S_NE
    #define RN_NW_N       RN_N_NW
    #define RN_NW_S       RN_S_NW

    #define RG_SE_NW      RG_NW_SE
    #define RG_SE_N       RG_N_SE
    #define RG_SE_NE      RG_NE_SE
    #define RG_SE_W       RG_W_SE
    #define RG_SE_E       RG_E_SE
    #define RG_SE_SW      RG_SW_SE
    #define RG_SE_S       RG_S_SE
    #define RG_S_N        RG_N_S
    #define RG_S_W        RG_W_S
    #define RG_S_E        RG_E_S
    #define RG_SW_NW      RG_NW_SW
    #define RG_SW_N       RG_N_SW
    #define RG_SW_NE      RG_NE_SW
    #define RG_SW_W       RG_W_SW
    #define RG_SW_S       RG_S_SW
    #define RG_E_NW       RG_NW_E
    #define RG_E_N        RG_N_E
    #define RG_E_W        RG_W_E
    #define RG_E_SW       RG_SW_E
    #define RG_W_NW       RG_NW_W
    #define RG_W_N        RG_N_W
    #define RG_NE_NW      RG_NW_NE
    #define RG_NE_N       RG_N_NE
    #define RG_NE_W       RG_W_NE
    #define RG_NE_E       RG_E_NE
    #define RG_NE_S       RG_S_NE
    #define RG_NW_N       RG_N_NW
    #define RG_NW_S       RG_S_NW

    #define BN_SE_NW      BN_NW_SE
    #define BN_SE_N       BN_N_SE
    #define BN_SE_NE      BN_NE_SE
    #define BN_SE_W       BN_W_SE
    #define BN_SE_E       BN_E_SE
    #define BN_SE_SW      BN_SW_SE
    #define BN_SE_S       BN_S_SE
    #define BN_S_N        BN_N_S
    #define BN_S_W        BN_W_S
    #define BN_S_E        BN_E_S
    #define BN_SW_NW      BN_NW_SW
    #define BN_SW_N       BN_N_SW
    #define BN_SW_NE      BN_NE_SW
    #define BN_SW_W       BN_W_SW
    #define BN_SW_S       BN_S_SW
    #define BN_E_NW       BN_NW_E
    #define BN_E_N        BN_N_E
    #define BN_E_W        BN_W_E
    #define BN_E_SW       BN_SW_E
    #define BN_W_NW       BN_NW_W
    #define BN_W_N        BN_N_W
    #define BN_NE_NW      BN_NW_NE
    #define BN_NE_N       BN_N_NE
    #define BN_NE_W       BN_W_NE
    #define BN_NE_E       BN_E_NE
    #define BN_NE_S       BN_S_NE
    #define BN_NW_N       BN_N_NW
    #define BN_NW_S       BN_S_NW

    #define BG_SE_NW      BG_NW_SE
    #define BG_SE_N       BG_N_SE
    #define BG_SE_NE      BG_NE_SE
    #define BG_SE_W       BG_W_SE
    #define BG_SE_E       BG_E_SE
    #define BG_SE_SW      BG_SW_SE
    #define BG_SE_S       BG_S_SE
    #define BG_S_N        BG_N_S
    #define BG_S_W        BG_W_S
    #define BG_S_E        BG_E_S
    #define BG_SW_NW      BG_NW_SW
    #define BG_SW_N       BG_N_SW
    #define BG_SW_NE      BG_NE_SW
    #define BG_SW_W       BG_W_SW
    #define BG_SW_S       BG_S_SW
    #define BG_E_NW       BG_NW_E
    #define BG_E_N        BG_N_E
    #define BG_E_W        BG_W_E
    #define BG_E_SW       BG_SW_E
    #define BG_W_NW       BG_NW_W
    #define BG_W_N        BG_N_W
    #define BG_NE_NW      BG_NW_NE
    #define BG_NE_N       BG_N_NE
    #define BG_NE_W       BG_W_NE
    #define BG_NE_E       BG_E_NE
    #define BG_NE_S       BG_S_NE
    #define BG_NW_N       BG_N_NW
    #define BG_NW_S       BG_S_NW

    #define YN_SE_NW      YN_NW_SE
    #define YN_SE_N       YN_N_SE
    #define YN_SE_NE      YN_NE_SE
    #define YN_SE_W       YN_W_SE
    #define YN_SE_E       YN_E_SE
    #define YN_SE_SW      YN_SW_SE
    #define YN_SE_S       YN_S_SE
    #define YN_S_N        YN_N_S
    #define YN_S_W        YN_W_S
    #define YN_S_E        YN_E_S
    #define YN_SW_NW      YN_NW_SW
    #define YN_SW_N       YN_N_SW
    #define YN_SW_NE      YN_NE_SW
    #define YN_SW_W       YN_W_SW
    #define YN_SW_S       YN_S_SW
    #define YN_E_NW       YN_NW_E
    #define YN_E_N        YN_N_E
    #define YN_E_W        YN_W_E
    #define YN_E_SW       YN_SW_E
    #define YN_W_NW       YN_NW_W
    #define YN_W_N        YN_N_W
    #define YN_NE_NW      YN_NW_NE
    #define YN_NE_N       YN_N_NE
    #define YN_NE_W       YN_W_NE
    #define YN_NE_E       YN_E_NE
    #define YN_NE_S       YN_S_NE
    #define YN_NW_N       YN_N_NW
    #define YN_NW_S       YN_S_NW

    #define YG_SE_NW      YG_NW_SE
    #define YG_SE_N       YG_N_SE
    #define YG_SE_NE      YG_NE_SE
    #define YG_SE_W       YG_W_SE
    #define YG_SE_E       YG_E_SE
    #define YG_SE_SW      YG_SW_SE
    #define YG_SE_S       YG_S_SE
    #define YG_S_N        YG_N_S
    #define YG_S_W        YG_W_S
    #define YG_S_E        YG_E_S
    #define YG_SW_NW      YG_NW_SW
    #define YG_SW_N       YG_N_SW
    #define YG_SW_NE      YG_NE_SW
    #define YG_SW_W       YG_W_SW
    #define YG_SW_S       YG_S_SW
    #define YG_E_NW       YG_NW_E
    #define YG_E_N        YG_N_E
    #define YG_E_W        YG_W_E
    #define YG_E_SW       YG_SW_E
    #define YG_W_NW       YG_NW_W
    #define YG_W_N        YG_N_W
    #define YG_NE_NW      YG_NW_NE
    #define YG_NE_N       YG_N_NE
    #define YG_NE_W       YG_W_NE
    #define YG_NE_E       YG_E_NE
    #define YG_NE_S       YG_S_NE
    #define YG_NW_N       YG_N_NW
    #define YG_NW_S       YG_S_NW

/* game status */
#define GAMEOVER        0
#define PLAYGAME        1
#define FIELDEDIT       2
#define TITLESCREEN     3

/* indicator locations */
#define ICONY           (-1)
#define CLOCKICON       (fieldx + 1)
#define MUSICICON       (-1)

/* control values */
#define NONE            0 /* must be 0 */
#define AMIGA           1
#define JOYSTICK        2
#define GAMEPAD         3
#define KEYBOARD        4
#define IBMPC           AMIGA
#define GBA             AMIGA
#define HUMAN           KEYBOARD

#define BANANAFRAMES      8 /* counting from 0 */
#define BIRDFRAMES        4 /* counting from 0 */
#define MISSILEFRAMES     5 /* counting from 0 */

#define PAIN_CREATURE     5
#define PAIN_HEAD         5
#define PAIN_PROTECTOR    5
#define PAIN_SLIME        2

/* 3. GLOBAL FUNCTIONS ---------------------------------------------------

engine1.c */

AGLOBAL void changefield(void);
AGLOBAL void clearhiscores(void);
AGLOBAL void enginesetup(void);
AGLOBAL void gameloop(void);
AGLOBAL void newfields(void);
AGLOBAL void newgame(void);
AGLOBAL void timeloop(void);
AGLOBAL void newfield(void);

AGLOBAL void drawsquare(SBYTE, SBYTE);
AGLOBAL void icon(SBYTE player, UBYTE image);
AGLOBAL void saylevel(COLOUR colour);
AGLOBAL void train(SCANCODE scancode, FLAG shift);
AGLOBAL void wormscore(SBYTE player, SLONG score);
AGLOBAL void align(STRPTR string, SBYTE size, TEXT filler);
AGLOBAL void wormcol(SBYTE player, SBYTE x, SBYTE y);
AGLOBAL void bombblast(UBYTE triggerer, SBYTE player, SBYTE centrex, SBYTE centrey, FLAG superbomb);
AGLOBAL void change(SBYTE x, SBYTE y, UBYTE image);

AGLOBAL SBYTE loadfields(STRPTR pathname);
AGLOBAL SBYTE partner(SBYTE which);
AGLOBAL FLAG savefields(STRPTR pathname);
AGLOBAL SBYTE valid(SBYTE x, SBYTE y);
AGLOBAL SBYTE xwrap(SBYTE x);
AGLOBAL SBYTE ywrap(SBYTE y);

/* system.c */

#ifdef __STORM__
    EXPORT int stci_d(char* out, int lvalue);
    EXPORT int stcl_d(char* out, long lvalue);
#endif

EXPORT void playsong(UBYTE whichsong);
/* EXPORT void cyclecolours(void); */
AGLOBAL void celebrate(void);
AGLOBAL void clearkybd(void);
AGLOBAL void clearscreen(void);
AGLOBAL void datestamp(void);
AGLOBAL void gameinput(void);
AGLOBAL void hiscores(void);
AGLOBAL void resettime(void);
AGLOBAL void systemsetup(void);
AGLOBAL void timing(void);
AGLOBAL void titlescreen(void);
AGLOBAL void turborender(void);
AGLOBAL void waitasec(void);
AGLOBAL void whiteline(void);

EXPORT void drawnumber(SBYTE x, UBYTE number);
EXPORT void rungame(void); // formerly timing()
EXPORT void drawscore(void);

AGLOBAL void cleanexit(SLONG rc);
AGLOBAL void effect(SBYTE); /* 'index' is a reserved word under Lint */
AGLOBAL void fileopen(FLAG revert);
AGLOBAL void filesaveas(FLAG flag);
AGLOBAL void flash(ULONG where);
AGLOBAL void systemrundown(SBYTE player, UBYTE multiply);
AGLOBAL void say(STRPTR sentence, COLOUR colour);
AGLOBAL void stopfx(void);
AGLOBAL void ReadAdapterJoystick(UBYTE unit);
AGLOBAL void ReadStandardJoystick(ULONG port);
AGLOBAL void ReadGamepads(void);
AGLOBAL void printstat(SBYTE player, SBYTE theline);
AGLOBAL void help(UBYTE type);
AGLOBAL void helpabout(void);
AGLOBAL void helploop(UBYTE type);
AGLOBAL void toggle(SBYTE key);
AGLOBAL void filedelete(void);
AGLOBAL void drawhead(SBYTE player);
AGLOBAL void colourline(SBYTE player);

AGLOBAL FLAG anykey(FLAG timeout);
AGLOBAL FLAG verify(void);
AGLOBAL FLAG ZOpen(STRPTR fieldname, FLAG write);
AGLOBAL FLAG ZRead(UBYTE* IOBuffer, ULONG length);
AGLOBAL FLAG ZWrite(UBYTE* IOBuffer, ULONG length);
AGLOBAL FLAG ZClose(void);
AGLOBAL ULONG wormobject(UBYTE player, SBYTE x, SBYTE y);

AGLOBAL void bangdynamite(SBYTE x, SBYTE y, SBYTE player);
AGLOBAL void checkrectangle(SBYTE direction, SBYTE player, SBYTE horizontalsize, SBYTE verticalsize);
AGLOBAL void putnumber(void);
AGLOBAL void turnworm(SBYTE player, SBYTE deltax, SBYTE deltay);
AGLOBAL void updatearrow(SBYTE arrowy);
AGLOBAL void protworm(SBYTE x, SBYTE y, UBYTE which1, UBYTE which2);
AGLOBAL void protprot(SBYTE x, SBYTE y, UBYTE which1, UBYTE which2);
AGLOBAL void squareblast(SBYTE type, SBYTE player, UBYTE c, SBYTE x, SBYTE y, FLAG cutter, FLAG superbomb);
AGLOBAL void drawmissile(SBYTE x, SBYTE y, UBYTE which);
AGLOBAL FLAG blockedtel(UBYTE which, SBYTE deltax, SBYTE deltay);
AGLOBAL SBYTE bsign(SBYTE value);
AGLOBAL FLAG  getnumber(SBYTE player);
AGLOBAL SBYTE speedup(SBYTE speed, FLAG isworm);
AGLOBAL ULONG arand(ULONG number);
AGLOBAL SBYTE whichteleport(SBYTE x, SBYTE y);
AGLOBAL void enginerundown(SBYTE player);
AGLOBAL void stat(SBYTE player, SBYTE line);
EXPORT void teleportcreature(SBYTE i, UBYTE which);
EXPORT SBYTE gettelx(SBYTE j);
EXPORT SBYTE gettely(SBYTE j);
EXPORT void printnumber(SBYTE x, UBYTE number);

/* engine2.c */
AGLOBAL void bothcol(SBYTE player, SBYTE x, SBYTE y);
AGLOBAL void createcreature(UBYTE species, UBYTE which, SBYTE x, SBYTE y, SBYTE deltax, SBYTE deltay, UBYTE player, UBYTE subspecies);
AGLOBAL void creatureloop(SBYTE which);
AGLOBAL void wormkillcreature(UBYTE player, UBYTE which);
AGLOBAL void protcreature(UBYTE player, UBYTE which);
AGLOBAL void wormcreature(UBYTE player, UBYTE which);
AGLOBAL void octopusfire(UBYTE which);
AGLOBAL UBYTE whichcreature(SBYTE x, SBYTE y, UBYTE species, UBYTE exception);
AGLOBAL void wormloop(SBYTE player);
AGLOBAL void wormqueue(SBYTE player, SBYTE deltax, SBYTE deltay);
AGLOBAL FLAG blockedsquare(SBYTE x, SBYTE y);
AGLOBAL void drawcreature(UBYTE which);
AGLOBAL void stoppedwormloop(SBYTE player);

/* fe.c */
AGLOBAL void fieldedit(void);
AGLOBAL void setbrush(UBYTE newbrush);

/* squares.c */
AGLOBAL void draw(SBYTE x, SBYTE y, UBYTE image);
AGLOBAL void fedraw(SBYTE x, SBYTE y, UBYTE image);

/* tails.c */
AGLOBAL void drawtail(SBYTE x, SBYTE y, UBYTE image);

/* heads.c */
AGLOBAL void realdrawhead(SBYTE x, SBYTE y, UBYTE image);

/* 4. GLOBAL STRUCTURES -------------------------------------------------- */

AGLOBAL struct WormStruct
{   SBYTE  x;
    SBYTE  y;
    SBYTE  deltax;
    SBYTE  deltay;
    SBYTE  position;
    SBYTE  posidir;
    SBYTE  statx;
    SBYTE  staty;
    SBYTE  pos;
    SBYTE  levelreached;
    SBYTE  arrowy;
    SBYTE  olddeltax;
    SBYTE  olddeltay;
    SBYTE  numbers;
    FLAG  affixer;
    FLAG  brakes;
    FLAG  encloser;
    FLAG  remnants;
    FLAG  alive;
    FLAG  moved;
    FLAG  pusher;
    FLAG  sideshot;
    FLAG  flashed;
    FLAG  rammed;
    FLAG  frosted;
    FLAG  wait;
    FLAG  autojump;
    UBYTE  speed;
    UBYTE  control;
    UBYTE  cause;
    UBYTE  last;
    UBYTE  multi;
    UBYTE  power;
    SWORD  armour;
    SWORD  ammo;
    SWORD  bias;
    SWORD  cutter;
    SWORD  lives;
    SWORD  glow;
    ULONG  score;
    ULONG  oldscore;
    ULONG  hiscore;
    ULONG  causewait;
    ULONG  nextlife;
    COLOUR colour;
    COLOUR dimcolour;
};
#define SPECIES 31
AGLOBAL struct CreatureInfoStruct
{   UBYTE  symbol;
    FLAG   wall; /* does it naturally live on walls? */
    UWORD  freq;
    UWORD  speed;
    ULONG  score;
};
