/* $Filename:    sys.cpp
 * $VER:         WormWars 8.22
 * $Description: Hardware/API calls for GBA
 *
 *  Copyright 2004-2009 James R. Jacobs.
 */

// too large for MULTIBOOT
// volatile const int __gba_multiboot = 1;

typedef unsigned short u16;
typedef unsigned long  u32;
#define REG_DISPCNT    *(u32*)0x4000000
#define MODE_4         0x4
#define BG2_ENABLE     0x400

/* #define TIME_FREQUENCY_1024 0x3
#define TIME_ENABLE         0x80
#define REG_TM3D            *(u16*)0x400010C
#define REG_TM3CNT          *(u16*)0x400010E */

#include "keypad.h"              // keypad definitions

#include "diff.h"
#include "same.h"
#include "logo.h"
#include "squares.h"
#include "heads.h"
#include "tails.h"

#include <stdlib.h>
#include <time.h>

// module functions
MODULE void PlotPixels(int x, int y, unsigned short int c);

// GLOBAL VARIABLES

EXPORT ULONG                DisplayWidth  = SCREENXPIXEL,
                            DisplayHeight = SCREENYPIXEL;
EXPORT FLAG                 paused;

// EXTERNAL VARIABLES -----------------------------------------------------

IMPORT FLAG                 engraved,
                            randomflag;
IMPORT SBYTE                a,
                            level, levels, players, reallevel,
                            sourcelevel;
IMPORT UBYTE                engraving,
                            field[MAXFIELDX + 1][MAXFIELDY + 1],
                            tailfield[MAXFIELDX + 1][MAXFIELDY + 1];
IMPORT SWORD                fieldx, fieldy,
                            secondsleft, secondsperlevel;
IMPORT ULONG                delay,
                            r,
                            quantity[5][3];
IMPORT struct WormStruct    worm[4];

// MODULE VARIABLES -------------------------------------------------------

MODULE COLOUR               blitter = NORMAL,
                            statcolour;
MODULE FLAG keydown[6]  = {FALSE, FALSE, FALSE, FALSE, FALSE, FALSE};
MODULE u16* paletteMem  = (u16*)0x5000000;
MODULE u16* videoBuffer = (u16*)0x6000000; // video memory
MODULE volatile unsigned short* ScanlineCounter = (volatile unsigned short*)0x4000006;

// CODE -------------------------------------------------------------------

int main(void)
{   // Start of program.

    REG_DISPCNT = MODE_4 | BG2_ENABLE; // switch to mode4

    randomflag      = FALSE;
    worm[0].control = NONE;
    worm[1].control = HUMAN;
    worm[2].control = NONE;
    worm[3].control = NONE;
    engraved        = TRUE;

    fieldx = STDFIELDX;
    fieldy = STDFIELDY;
    
    enginesetup();
    newfields();

    for (;;)
    {   titlescreen();
        newgame();
}   }

// SUPPORT FUNCTIONS -----------------------------------------------------

EXPORT void fileabout(void)                             { ; }
EXPORT void freefx(void)                                { ; }
EXPORT void pause(void)                                 { ; }
EXPORT void unpause(void)                               { ; }
EXPORT void effect(SBYTE index)                         { ; }
EXPORT void systemsetup(void)                           { ; }
EXPORT void toggle(SBYTE key)                           { ; }
EXPORT void next(void)                                  { ; }
EXPORT void cleartext(void)                             { ; }
EXPORT void datestamp(void)                             { ; }
EXPORT void flash(ULONG where)                          { ; }
EXPORT void stopfx(void)                                { ; }
EXPORT void playsong(UBYTE whichsong)                   { ; }
EXPORT void domusic(void)                               { ; }
EXPORT void wwhelp(SBYTE type)                          { ; }
EXPORT void waitasec(void)                              { ; }
EXPORT void icons(SBYTE player)                         { ; }	
EXPORT void systemrundown(SBYTE player, UBYTE multiply) { ; }
EXPORT void say(STRPTR sentence, COLOUR colour)         { ; }
EXPORT void printstat(SBYTE player, SBYTE theline)      { ; }
EXPORT void resettime(void)                             { ; }
EXPORT void stci_d(void)                                { ; }
EXPORT void stcl_d(void)                                { ; }
EXPORT void cleanexit(SLONG rc)                         { ; }
EXPORT void clearscreen(void)                           { ; }

EXPORT void clearkybd(void)
{   int player;

    for (player = 0; player <= 3; player++)
    {   worm[player].pos = -1;
}   }

EXPORT void celebrate(void)
{   a = GAMEOVER;
}

EXPORT void gameinput(void)
{   if (!((*KEYS) & KEY_SELECT))
    {   if (!keydown[5])
        {   if (paused)
            {   paused = FALSE;
                draw(19, 12, BLACKENED);
            } else
            {   paused = TRUE;
                draw(19, 12, CLOCK);
        }   }
        keydown[5] = TRUE;
    } else keydown[5] = FALSE;

    if (paused) return;

    if (!((*KEYS) & KEY_A))
    {   if (!keydown[0])
        {   wormqueue(1,  0,  0);
        }
        keydown[0] = TRUE;
    } elif (!((*KEYS) & KEY_B))
    {   if (!keydown[0])
        {   wormqueue(1,  0,  0);
        }
        keydown[0] = TRUE;
    } else keydown[0] = FALSE;

    if (!((*KEYS) & KEY_LEFT))
    {   if (!keydown[1])
        {   wormqueue(1, -1,  0);
        }
        keydown[1] = TRUE;
    } else keydown[1] = FALSE;

    if (!((*KEYS) & KEY_RIGHT))
    {   if (!keydown[2])
        {   wormqueue(1, +1,  0);
        }
        keydown[2] = TRUE;
    } else keydown[2] = FALSE;

    if (!((*KEYS) & KEY_UP))
    {   if (!keydown[3])
        {   wormqueue(1,  0, -1);
        }
        keydown[3] = TRUE;
    } else keydown[3] = FALSE;

    if (!((*KEYS) & KEY_DOWN))
    {   if (!keydown[4])
        {   wormqueue(1,  0, +1);
        }
        keydown[4] = TRUE;
    } else keydown[4] = FALSE;
}

EXPORT void blitmode(COLOUR state)
{   blitter = state;
}

EXPORT void draw(SBYTE x, SBYTE y, UBYTE image)
{   SWORD sourcex, sourcey, destx, desty, xtimes;

    if (blitter == BLACK)
        image = BLACKENED;
    elif (blitter == WHITE)
        image = WHITENED;
    elif (image == EMPTY && engraved)
        image = FIRSTPATTERN + engraving;

    sourcex = (image % 20) * (SQUAREX / 2);
    sourcey = (image / 20) * SQUAREY;
    for
    (   desty = y * SQUAREY;
        desty < (y * SQUAREY) + SQUAREY;
        desty++, sourcey++
    )
    {   xtimes = 0;
        for
        (   destx = (x * SQUAREX) / 2;
            xtimes < SQUAREX;
            xtimes++
        )
        {   PlotPixels(destx + (xtimes / 2), desty, squaresData[sourcey * 120 + (sourcex + (xtimes / 2))]);
}   }   }

EXPORT void titlescreen(void)
{   TRANSIENT int  i,
                   x, y;
    PERSIST   FLAG first = TRUE;

    a = TITLESCREEN;

    for (i = 0; i < 256; i++) // copy palette to memory
        paletteMem[i] = picPalette[i];

    /* draw the picture */
    for (y = 0; y < 160; y++)
    {   for (x = 0; x < 120; x++)
        {   PlotPixels(x, y, picData[(y * 120) + x]);
    }   }
    if (first)
    {   first = FALSE;
    } else
    {   drawscore();
    }

    while (a == TITLESCREEN)
    {   if (!((*KEYS) & KEY_A) || !((*KEYS) & KEY_B))
        {   keydown[0] = TRUE;
            a = PLAYGAME;
        } elif (!((*KEYS) & KEY_START))
        {   a = PLAYGAME;
}   }   }

EXPORT void rungame(void)
{   int i,
        x;

    srand((unsigned) time(NULL));
    paused = FALSE;

    for (i = 0; i < 256; i++) // copy palette to memory
    {   paletteMem[i] = squaresPalette[i];
    }
    for (x = 0; x < 114; x++)
    {   PlotPixels(x, 144, (109 << 8) + 109);
    }

    while (a == PLAYGAME)
    {   r++;
        gameloop();

        while (*ScanlineCounter <  160) { ; } // ie. WaitVBlank();
        while (*ScanlineCounter >= 160) { ; }
        while (*ScanlineCounter <  160) { ; } // ie. WaitVBlank();
        while (*ScanlineCounter >= 160) { ; }
        while (*ScanlineCounter <  160) { ; } // ie. WaitVBlank();
        while (*ScanlineCounter >= 160) { ; }
        while (*ScanlineCounter <  160) { ; } // ie. WaitVBlank();
}   }

void turborender(void)
{   SBYTE x, y;

    changefield();
    for (x = 0; x <= fieldx; x++)
    {   for (y = 0; y <= fieldy; y++)
        {   draw(x, y, field[x][y]);
    }	}
    for (y = 0; y <= fieldy; y++)
    {   draw(fieldx + 1, y, BLACKARROW);
}   }

EXPORT void drawtail(SBYTE x, SBYTE y, UBYTE image)
{   SWORD sourcex, sourcey, destx, desty, xtimes;

    sourcex = (image % 20) * (SQUAREX / 2);
    sourcey = (image / 20) * SQUAREY;
    for
    (   desty = y * SQUAREY;
        desty < (y * SQUAREY) + SQUAREY;
        desty++, sourcey++
    )
    {   xtimes = 0;
        for
        (   destx = (x * SQUAREX) / 2;
            xtimes < SQUAREX;
            xtimes++
        )
        {   PlotPixels(destx + (xtimes / 2), desty, tailData[sourcey * 120 + (sourcex + (xtimes / 2))]);
    }   }

    tailfield[x][y] = (UBYTE) (image % SECONDMAGIC); // decolourized
}
EXPORT void realdrawhead(SBYTE x, SBYTE y, UBYTE image)
{   SWORD sourcex, sourcey, destx, desty, xtimes;

    sourcex = (image % 20) * (SQUAREX / 2);
    sourcey = (image / 20) * SQUAREY;
    for
    (   desty = y * SQUAREY;
        desty < (y * SQUAREY) + SQUAREY;
        desty++, sourcey++
    )
    {   xtimes = 0;
        for
        (   destx = (x * SQUAREX) / 2;
            xtimes < SQUAREX;
            xtimes++
        )
        {   PlotPixels(destx + (xtimes / 2), desty, headData[sourcey * 120 + (sourcex + (xtimes / 2))]);
}   }   }

EXPORT void whiteline(void)
{   statcolour = WHITE;
}
EXPORT void colourline(SBYTE player)
{   statcolour = worm[player].colour;
}

MODULE void PlotPixels(int x, int y, unsigned short int c)
{   videoBuffer[(y) * 120 + (x)] = (c);
}

EXPORT FLAG busyloop(FLAG any)
{   return FALSE;
}
EXPORT FLAG anykey(FLAG timeout)
{   return FALSE;
}
EXPORT FLAG ZOpen(STRPTR fieldname, FLAG write)
{   return TRUE;
}
EXPORT FLAG ZRead(UBYTE* IOBuffer, ULONG length)
{   return TRUE;
}
EXPORT FLAG ZWrite(UBYTE* IOBuffer, ULONG length)
{   return TRUE;
}
EXPORT FLAG ZClose(void)
{   return TRUE;
}
EXPORT FLAG verify(void)
{   return(TRUE);
}

EXPORT void drawscore(void)
{   int score;

    drawnumber(0, worm[1].score / 1000000);
    score = worm[1].score % 1000000;
    drawnumber(1, score /  100000);
    score %=  100000;
    drawnumber(2, score /   10000);
    score %=   10000;
    drawnumber(3, score /    1000);
    score %=    1000;
    drawnumber(4, score /     100);
    score %=     100;
    drawnumber(5, score /      10);
    score %=      10;
    drawnumber(6, score          );
}

EXPORT void drawnumber(SBYTE x, UBYTE number)
{   SWORD sourcey, destx, desty, xtimes;

    sourcey = 155 - (number * 6);
    for
    (   desty = 155;
        desty < 160;
        desty++, sourcey++
    )
    {   xtimes = 0;
        for
        (   destx = x * 6 / 2; // should be * 5 not * 6 really
            xtimes < 6; // should be 5 not 6 really
            xtimes++
        )
        {   PlotPixels(destx + (xtimes / 2), desty, headData[sourcey * 120 + (xtimes / 2)]);
}   }   }

